/*
 *	Copyright (c) Vanderbilt University, 2001-2002
 *	ALL RIGHTS RESERVED
 *
 *	Vanderbilt University disclaims all warranties with regard to this
 *	software, including all implied warranties of merchantability
 *	and fitness.  In no event shall Vanderbilt University be liable for
 *	any special, indirect or consequential damages or any damages
 *	whatsoever resulting from loss of use, data or profits, whether
 *	in an action of contract, negligence or other tortious action,
 *	arising out of or in connection with the use or performance of
 *	this software.
 *
 *  FILE NAME:	PatternMatcher.cpp 
 *	SYNOPSIS:	Pattern Match Interface Implementation
 *  AUTHOR:		Feng Shi - feng.shi@vanderbilt.edu
 *              Tihamr Levendovszky - tihamer.levendovszky@vanderbilt.edu
 *  CREATED:	 06/11/02
 *  LAST UPDATE: 06/24/02
 *  CHANGE LOG:

		10/07/02  - Feng

		For AssociationClass: A-->Connection/AssociationClass-->B
		Treat the AssociationClass as other PatternObject
		Consider the case: A or B or both must be bound, AssociationClass(AC) is not bound.
		But how about other bound case: AC bound, A & B not bound, or A or B bound
 
		SimpleAssociation: Reference, Set Membership, Archetype--Instance
 */

#pragma warning ( disable : 4786 ) // Names more than 255 characters


#include <Windows.h>

#include <sstream>

#include "PatternMatcher.h"
#include "GReATUtils.h"

//#define DEBUG
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
PatternMatcher::~PatternMatcher(){}

void PatternMatcher::findMatch()
{
	if((string)thisRule.name() == "Rule")
	{
		if(!init()) return;		
		reqRunMatch();
		return;
	}
	if(!init()) return;		
	reqRunMatch();
}

bool PatternMatcher::init()
{
//Step 1: Check the GR object with non-zero objectID
	set<GR::ObjectWrapper> grObjs = thisRule.ObjectWrapper_kind_children();
	for(set<GR::ObjectWrapper>::const_iterator i=grObjs.begin();i!=grObjs.end();++i)
	{
		GR::ObjectWrapper grObj = *i;
		Object object;
		if(grObj.objectID()!=0 && grObj.dataNetworkID()!=0) 
		{
			if(!backendMap->GetObject(grObj, object)) return false;
		}
		else
		{//check whether it is global object			
			string gname = (string)grObj.paradigmName()+"::"+(string)grObj.classType();
			if(!backendMap->GetGlobalObject(gname, object)) continue;
		}
		// Check the negative condition before binding
		if(!checkNegative(grObj, object)) return false;
		bind(grObj, object);
	}

// 10-09-02:
//	Check the validity of all patternLinks if its two ends both bound
	set<GR::ObjectLink> grLinks = thisRule.ObjectLink_kind_children();
	for(set<GR::ObjectLink>::const_iterator j=grLinks.begin();j!=grLinks.end();++j)
	{
		GR::ObjectLink currPl = *j;		

		if((string)currPl.cardinality()=="0") continue;		
		if((string)currPl.action()==(string)"CreateNew") continue;
		if(isBound(currPl)) continue;   // for the situation of another Link of AssociationClass		
		
		GR::ObjectWrapper srcObj = currPl.src_end();
		GR::ObjectWrapper dstObj = currPl.dst_end();
		if(srcObj.objectID()==0 || dstObj.objectID()==0) continue;
		
		Object obSrc, obDst;
		if(!isValidLink(currPl,obSrc,obDst)) return false;
	}
	
	return true;
}

bool PatternMatcher::getEdgeToBound(GR::ObjectWrapper &boundPO,GR::ObjectWrapper &unboundPO, \
									GR::ObjectLink &unboundPl)
{
	set<GR::ObjectLink> objLinks=thisRule.ObjectLink_kind_children();
	for(set<GR::ObjectLink>::const_iterator i=objLinks.begin();i!=objLinks.end();++i)
	{
		unboundPl = *i;
		if(isBound(unboundPl)) continue;

		if((string)unboundPl.cardinality()=="0") continue;
		if((string)unboundPl.action()==(string)"CreateNew") continue;

		GR::ObjectWrapper srcPO = unboundPl.src_end();
		GR::ObjectWrapper dstPO = unboundPl.dst_end();
		
		if(srcPO.objectID()==0 && dstPO.objectID()==0) continue;
		
		if(srcPO.objectID()==0 && dstPO.objectID()!=0)
		{
			boundPO=dstPO;
			unboundPO=srcPO;
			return true;
		}
		if(srcPO.objectID()!=0 && dstPO.objectID()==0)
		{
			boundPO=srcPO;
			unboundPO=dstPO;
			return true;
		}
	}	
	return false;
}

bool PatternMatcher::reqRunMatch()
{
/*
	This function will be called recursively.
*/
	if(!thisRule.exhaustive() && matches.size()==1) return true;

	GR::ObjectWrapper boundPO;
	GR::ObjectWrapper unboundPO;
	GR::ObjectLink unboundPl;

	if(getEdgeToBound(boundPO, unboundPO, unboundPl))
	{
		Object patternObject;
		backendMap->GetObject(boundPO, patternObject);

#ifdef DEBUG
		cout<<endl<<"Bound Object: "<<string(boundPO.name())<<"("<<bname<<")  UnBound Object: "\
			<<string(unboundPO.name())<<endl;
#endif
		
		GR::ObjectWrapper poSrcTemp = unboundPl.src_end();
		GR::ObjectWrapper poDstTemp = unboundPl.dst_end();
			
		const string &srcType = poSrcTemp.classType();
		const string &dstType = poDstTemp.classType();
		Uml::Class srcClsType, dstClsType;
		if(!backendMap->GetUmlClass(poSrcTemp,srcClsType) \
			|| !backendMap->GetUmlClass(poDstTemp, dstClsType))
		{
			throw udm_exception("No existing class "+srcType+" or "+dstType);
		}	
			
		bool isSrc=(poSrcTemp.objectID()!=0);
		set<Object> adjHostVertices;
		backendMap->GetAdjacentObjects(unboundPl,patternObject,isSrc, adjHostVertices);

#ifdef DEBUG				
		if(adjHostVertices.empty()) cout<<"adjHostVertices is empty!"<<endl;
		else cout<<"adjHostVertices size: "<<adjHostVertices.size()<<endl;
#endif

		string unboundObjMetaName;
		if(isSrc) unboundObjMetaName=poDstTemp.paradigmName();
		else unboundObjMetaName=poSrcTemp.paradigmName();

		// Iterate through the adjacent host vertices
		for(set<Object>::const_iterator ii=adjHostVertices.begin();ii!=adjHostVertices.end();++ii)
		{
			Object adjObject = *ii;
//			if(isBound(adjObject,unboundObjMetaName)) continue;  //In example ESML2Config, Processor-->Process-->NormalComponent
			if(isBound(adjObject)) continue;

			Uml::Class clsPatternObject;
			if(!backendMap->GetUmlClass(unboundPO,clsPatternObject))
			{
				cerr<<endl<<"Error: Non existing class "<<string(unboundPO.classType())\
					<<" has been specified in Pattern Object "<<string(unboundPO.name())<<"."<<endl;
				continue;
			}

			const Uml::Class& clsObject=adjObject.type();
			if(!Uml::IsDerivedFrom(clsObject,clsPatternObject)) continue;

#ifdef DEBUG
			cout<<"adjObject name: "<<extractName(adjObject)\
				<<"-----Unbound PatternObject: "<<string(unboundPO.name())<<endl;
			cout<<"patternObject name: "<<extractName(patternObject)\
				<<"-----Bound PatternObject: "<<string(boundPO.name())<<endl;
#endif
		// Check the negative condition before binding
			if(!checkNegative(unboundPO, adjObject)) continue;

			bind(unboundPO, adjObject);  
			bind(unboundPl);

			if(bindParallelEdges(unboundPO))
			{
				reqRunMatch();
			}
					
			unBind(unboundPO, adjObject);
		}	
	}
	else 
	{
		if(executeGuard() && isAllBound()) matches.push_back(currentMatch);
		return true;
	}
	
	return false;
}

bool PatternMatcher::isAllBound()
{
	set<GR::ObjectLink> grLinks = thisRule.ObjectLink_kind_children();
	for(set<GR::ObjectLink>::const_iterator j=grLinks.begin();j!=grLinks.end();++j)
	{
		GR::ObjectLink currPl = *j;		

		if((string)currPl.cardinality()=="0") continue;		
		if((string)currPl.action()==(string)"CreateNew") continue;
		if(!isBound(currPl)) return false;
	}
	
	return true;
}

bool PatternMatcher::bindParallelEdges(GR::ObjectWrapper &poNew)
{
	// If there is an unbound pattern edge in the pattern
	// between the newly bound pattern object (po) and the
	// other already bound PatternObjects, then check these
	// unbound edges and return false: the newly bound edge
	// does not match the pattern
	
	set<GR::ObjectLink> plDst=poNew.dst();
	set<GR::ObjectLink> plSrc=poNew.src();

	set<GR::ObjectLink> plFromPo;
	plFromPo.insert(plSrc.begin(), plSrc.end());
	plFromPo.insert(plDst.begin(), plDst.end());

	return reqBindEdgeList(plFromPo,poNew);
}

bool PatternMatcher::reqBindEdgeList(set<GR::ObjectLink>& plList,GR::ObjectWrapper& poNew)
{	
	// Picking one pattern link from the list
	if(plList.empty()) return true;
	GR::ObjectLink plCurr=*(plList.begin());
	plList.erase(plList.begin());
	
	// Bound edges are OK.
	if(isBound(plCurr) || (string)plCurr.action()=="CreateNew" || (string)plCurr.cardinality()=="0")
	{
		return reqBindEdgeList(plList,poNew);
	}

	// Getting the adjacent PatternObject
	GR::ObjectWrapper poDst = plCurr.dst_end();
	GR::ObjectWrapper poSrc = plCurr.src_end();
	GR::ObjectWrapper poAdj = ( poDst==poNew ? poSrc : poDst);

	// To see what are the participants here
#ifdef DEBUG
	string poDstName=poDst.name(); cout<<poDstName<<"  "<<poDst.objectID()<<endl;
	string poSrcName=poSrc.name(); cout<<poSrcName<<"  "<<poSrc.objectID()<<endl;
	string poAdjName=poAdj.name(); cout<<poAdjName<<"  "<<poAdj.objectID()<<endl;
	string poNewName=poNew.name(); cout<<poNewName<<"  "<<poNew.objectID()<<endl;
#endif

	if(poAdj.objectID()==0)
	{
		return reqBindEdgeList(plList,poNew);
	}

	Object obSrc, obDst;
	if(!isValidLink(plCurr,obSrc,obDst)) return false;

	if(!reqBindEdgeList(plList,poNew)) // If cannot find match
	{
		unBind(plCurr);
		return false;
	}

	return true;
}

bool PatternMatcher::isValidLink(GR::ObjectLink& objLink,Object& obSrc,Object& obDst)
{
	GR::ObjectWrapper poDst = objLink.dst_end();
	GR::ObjectWrapper poSrc = objLink.src_end();

	backendMap->GetObject(poSrc, obSrc);
	backendMap->GetObject(poDst, obDst);
	
	set<Object> adjObjs;
	backendMap->GetAdjacentObjects(objLink,obSrc,true,adjObjs);
	set<Udm::Object>::iterator pos = adjObjs.find(obDst);
	if(pos==adjObjs.end()) return false;
	bind(objLink);
	return true;
}

bool PatternMatcher::checkNegative(const GR::ObjectWrapper &grObj, Object &object) const
{
	set<GR::ObjectLink> srcLinks = grObj.src();
	set<GR::ObjectLink> dstLinks = grObj.dst();
	set<GR::ObjectLink> links;
	links.insert(srcLinks.begin(), srcLinks.end());
	links.insert(dstLinks.begin(), dstLinks.end());
	for(set<GR::ObjectLink>::const_iterator i=links.begin();i!=links.end();++i)
	{
		GR::ObjectLink currLink = *i;
		if((string)currLink.cardinality()!=(string)"0") continue;
		GR::ObjectWrapper srcObj = currLink.src_end();
		GR::ObjectWrapper dstObj = currLink.dst_end();
		
		set<Object> adjObjs;
		if(srcObj==grObj) 
		{
			backendMap->GetAdjacentObjects(currLink,object,true,adjObjs);
		}
		else 
		{
			backendMap->GetAdjacentObjects(currLink,object,false,adjObjs);
		}
		if(!adjObjs.empty()) 
		{
			return false;
		}
	}
	return true;
}

bool PatternMatcher::isBound(const Object& object, const string& metaName) const
{		
	map<string, Object>::const_iterator i=currentMatch.boundObjMap.find(getUId(metaName, object));
	if(i!=currentMatch.boundObjMap.end()) return true;
	else return false;
}

bool PatternMatcher::isBound(const Object& object) const
{
	unsigned long dnid = object.__impl()->__getdn()->uniqueId();
	#if _MSC_VER >= 1310
		set< pair<unsigned long, Object> >::const_iterator i=currentMatch.boundObject.find(pair<unsigned long, Object>(dnid, object));
	#else
		set< pair<unsigned long, Object> >::iterator i=currentMatch.boundObject.find(pair<unsigned long, Object>(dnid, object));
	#endif

	if(i!=currentMatch.boundObject.end()) return true;
	else return false;
}

bool PatternMatcher::isBound(const GR::ObjectLink& objLink) const
{
	#if _MSC_VER >= 1310
		set<GR::ObjectLink>::const_iterator i=currentMatch.boundGRLink.find(objLink);
	#else
		set<GR::ObjectLink>::iterator i=currentMatch.boundGRLink.find(objLink);
	#endif

	if(i!=currentMatch.boundGRLink.end()) return true;
	else return false;
}


void PatternMatcher::bind(GR::ObjectWrapper& grObj,const Object& object)
{
	const string& clsname=grObj.classType();
	Uml::Class clsFound;

	if(backendMap->GetUmlClass(grObj, clsFound))
	{	
		grObj.objectID() = object.uniqueId();
		currentMatch.grObj2BoundObj[grObj] = object;
//		currentMatch.boundObjMap[getUId((string)grObj.paradigmName(), object)] = object;
		unsigned long dnid = object.__impl()->__getdn()->uniqueId();
		grObj.dataNetworkID() = dnid;
		currentMatch.boundObject.insert(pair<unsigned long, Object>(dnid, object));
	}
	else 
	{
		throw udm_exception("No existing class "+clsname);
	}
}

void PatternMatcher::bind(GR::ObjectLink& objLink)
{
	currentMatch.boundGRLink.insert(objLink);
}

void PatternMatcher::unBind(GR::ObjectWrapper& grObj,Object& object)
{
// First remove BoundLinks from all BoundObject	
	set<GR::ObjectLink> dstLinks=grObj.dst();
	set<GR::ObjectLink> srcLinks=grObj.src();

	set<GR::ObjectLink> grLinks;
	grLinks.insert(dstLinks.begin(),dstLinks.end());
	grLinks.insert(srcLinks.begin(),srcLinks.end());

	for(set<GR::ObjectLink>::const_iterator j=grLinks.begin();j!=grLinks.end();++j)
	{
		GR::ObjectLink grLink = *j;
		unBind(grLink);
	}

// Second, remove GRObject from grObj2BoundObj
	map<GR::ObjectWrapper, Object>::iterator i=currentMatch.grObj2BoundObj.find(grObj);
	if(i!=currentMatch.grObj2BoundObj.end()) currentMatch.grObj2BoundObj.erase(i);

//Third, remove from boundObjMap
//	map<string, Object>::iterator ii=currentMatch.boundObjMap.find(getUId((string)grObj.paradigmName(), object));
//	if(ii!=currentMatch.boundObjMap.end()) currentMatch.boundObjMap.erase(ii);

	unsigned long dnid = object.__impl()->__getdn()->uniqueId();
	set< pair<unsigned long, Object> >::iterator ii=currentMatch.boundObject.find(pair<unsigned long, Object>(dnid, object));
	if(ii!=currentMatch.boundObject.end()) currentMatch.boundObject.erase(ii);

	grObj.objectID() = 0;
	grObj.dataNetworkID() = 0;
}

void PatternMatcher::unBind(GR::ObjectLink& grLink)
{
	set<GR::ObjectLink>::iterator j=currentMatch.boundGRLink.find(grLink);
	if(j!=currentMatch.boundGRLink.end()) currentMatch.boundGRLink.erase(j);
}

/*void PatternMatcher::invokeCint()
{
	
	
}
*/
extern HINSTANCE hCINTDLL;
typedef bool (*MAP_FN)(const vector <Udm::Object> &objs, bool& err);
bool PatternMatcher::executeGuard()
{
	string guardStr = thisRule.guard();
	if(guardStr.empty()) return true;

	ostringstream gFStream;
	const __int64 &o_id = thisRule.ruleId();
	gFStream << "Guard" << "_" << convertToCPlusPlusName((string)thisRule.name()) << "_" << o_id ;
	string fn = gFStream.str();
	//cout << "Executing Guard function " << fn << endl;

	MAP_FN lpMapFn = (MAP_FN)GetProcAddress(hCINTDLL, fn.c_str());

	if(lpMapFn == NULL)
	{	cout << "ERROR: Guard function " << fn << " not found in dll \n";
		return true;
	}

	vector <Udm::Object > objs;
	set<string > names;
//	const vector<GR::PatternObject> &patobjs = thisRule.PatternObject_kind_children();
//	vector<GR::PatternObject>::const_iterator i;
	string name;

	typedef NameCompareFunctor< GR::PatternObject> PatternObjectCompareFunc;
	typedef set< GR::PatternObject, PatternObjectCompareFunc> PattObjSorted_t;
	PattObjSorted_t patobjs = thisRule.PatternObject_kind_children_sorted( PatternObjectCompareFunc());
	PattObjSorted_t::const_iterator i;

	for(i=patobjs.begin();i!=patobjs.end();i++)
	{
		const GR::PatternObject &po = *i;
		if((string)po.action() == "CreateNew") continue;

		name = po.name();
		set<string>::iterator ni = names.find(name);
		if(ni!=names.end())
			continue;
		names.insert(name);
//		const Udm::Object &o = currentMatch.grObj2BoundObj[po]; //Bug here, for some po not in the map, operation[] will insert it
//		objs.push_back(o);
		Udm::Object o;
		if(currentMatch.grObj2BoundObj.find(po)!=currentMatch.grObj2BoundObj.end())
		{
			o = currentMatch.grObj2BoundObj[po];
		}
		objs.push_back(o);
	}


  bool err = false;
	bool ret = false;
	ret = lpMapFn(objs,err);
  if (err)
    cout << "ERROR: Guard function " << fn << " could not execute\n";

	return ret;
}

string PatternMatcher::getUId(const string& metaName, const Object& obj) const 
{
	string uid(metaName);
	uid+="::";
	char buffer[20];
	_ltoa(obj.uniqueId(), buffer, 10);
	uid += buffer;
	return uid;

	//need add the datanetwork id here 
}