#include "ConfigGenerator.h"

#include <fstream>

#include <UdmBase.h>
#include <Uml.h>

#include "GReATConfig.h"
#include "GReATConfigRuntime.h"

#include "CreateDNCodePrinter.h"

#include "../CodeGenerator/PE.h"
#include "../Codegenerator/PEStringTable.h"
#include "../CodeGenerator/CGEx.h"
#include "GReATUtils.h"
#include "GRCWrapper.h"

///////////////////////////////////////////////////////////////////////////////
std::string createTempRootNames( const GReATConfig::TemporaryFileUsage& tfu) {
	return ( std::string)tfu.MetaName() + RES+ ( std::string)tfu.RootClassName();
}

///////////////////////////////////////////////////////////////////////////////
ConfigGenerator::ConfigGenerator( const std::string& configFileName) : _cic( 0), _clc( 0), _cg( 0), _cfg_fn( configFileName)
{
	// open config dn & get the root.
	_sdnConfig= new Udm::SmartDataNetwork( GReATConfig::diagram);
	_sdnConfig->OpenExisting( _cfg_fn, "GReATConfig", Udm::CHANGES_LOST_DEFAULT);
	GReATConfig::RootFolder root= GReATConfig::RootFolder::Cast( _sdnConfig->GetRootObject());
	// gain access to meta & usage info
	_metaInfo= root.MetaInformation_child();
	_usageInfo= root.UsageInformation_child();
	if ( Udm::null == _metaInfo) // meta is compulsory
		throw CGEx( "FATAL: No meta information found in the configuration file: " + _cfg_fn);
	_metaConfig= new GReATConfigMeta( _metaInfo);
	_usageConfig= new GReATConfigUsage( _usageInfo);
	// add meta diagrams
	GeneralBackend::BackendMap_t backendMap;
	_metaConfig->getGeneralBackend( backendMap, getPath( _cfg_fn));
	_backend= backendMap;
	// add cross link diagram, if neccessary.
	if( _metaConfig->isUdmProjectFile())
		_backend.setCrossLinkBackend( _metaConfig->getCrossLinkDiagramName(), _metaConfig->getCrossLinkDN());
	// decide if am, guard codes should be printed into the generated code, or called from a precompiled library
	GReATConfigMeta::MetaCodeGenModes mcgm= _metaConfig->getMetaCodeGenMode();
	ICIntCompiler* icc= 0;
	if ( GReATConfigMeta::isLibraryToGenerate( mcgm)) {
		icc= _clc= new CIntLibraryCall;
		generateUCL( "", mcgm);
		_gzInc= dropExtension( dropPath( _metaConfig->getTransformationFileName()));
	} else if ( mcgm == GReATConfigMeta::MCGM_NO_LIB) {
#if _MSC_VER < 1300
		generateSpecCmpCode( "", mcgm);
#endif
		icc= _cic= new CIntCompiler;
	} else {
		throw CGEx("Invalid meta code generation mode found in configuration.");
	}
	// collect temporary roots
	std::vector< GReATConfig::TemporaryFileUsage> tempFiles;
	_usageConfig->getTemporaryFileUsages( tempFiles);
	std::list< std::string> tempRoots;
	std::transform( tempFiles.begin(), tempFiles.end(), std::back_inserter( tempRoots), &createTempRootNames);
	_backend.generateTempRoots( tempRoots);
	//
	_cg= new CodeGenerator( _backend, *icc, _gzInc);
}

ConfigGenerator::~ConfigGenerator()
{
	delete _cic;
	_cic= NULL;
	delete _clc;
	_clc= NULL;
	delete _cg;
	_cg= NULL;
	delete _sdnConfig;
	_sdnConfig= NULL;
	delete _metaConfig;
	_metaConfig= NULL;
	delete _usageConfig;
	_usageConfig= NULL;
}

void ConfigGenerator::generate()
{
	GR::RootFolder grRoot;
	_metaConfig->getRootFolder( grRoot, getPath( _cfg_fn));
	_cg->generate( grRoot);
}

void ConfigGenerator::generateMain()
{
	if ( Udm::null == _usageInfo)
		throw CGEx( "FATAL: No usage information found in the configuration file.");
	// main skeleton
	createTempFileUsageInfos();
	std::vector< std::string> fileIDs;
	_usageConfig->getFileUsageFileIDs( fileIDs);
	PEComposite* output;
	createExternGlobalDefTemporaryRootObject( _main);
	generateMainSkeleton(  fileIDs, &output);
	// Generate (but not add to output yet) the code which init, open and close datanetworks.
	generateDNCode( fileIDs, *output);
	// init&open code for DNs
	createOpenDNCode( *output);
	// get start rule information
	GR::RewritingRule startRule;
	getStartRule( startRule);
	const Class* startRuleClass= _cg->_translator.ruleDef( startRule);
	const ParamDatas_t* ruleInputs= _cg->_passPacket.inputParams( startRule);
	const ParamDatas_t* ruleOutputs= _cg->_passPacket.outputParams( startRule);
	// add start rule packet definitions to 'output'.
	createPacketDefinitions( *ruleInputs, *output);
	createPacketDefinitions( *ruleOutputs, *output);
	// add 'get objects from DNs' code to 'output'
	createGetObjectsFromDNs( *ruleInputs, *output);
	createGetTemporaryRootObject( *output);
	// transformation
	createExecuteTransformation( *startRuleClass, *ruleInputs, *ruleOutputs, *output);
	// close code for DNs
	createCloseDNCode( *output);
}

void ConfigGenerator::saveVisualStudioProject( const std::string& name) {
	if ( !_metaConfig) {
		throw CGEx("Code generator internal error. ConfigGenerator::MetaConfig is null.");
	}
//	std::string currDir= getPath( name);
	std::string cfgPath= getPath( _cfg_fn);
	GRCWrapper().createVSProj( name, *_metaConfig, cfgPath.empty() ? getCwd() : cfgPath, _metaConfig->getMetaCodeGenMode());
}

void ConfigGenerator::generateExe( const std::string& name) {
	if ( !_metaConfig) {
		throw CGEx("Code generator internal error. ConfigGenerator::MetaConfig is null.");
	}
	std::string cfgPath= getPath( _cfg_fn);
	GRCWrapper().createExe( name, *_metaConfig, cfgPath.empty() ? getCwd() : cfgPath, _metaConfig->getMetaCodeGenMode());
}

void ConfigGenerator::save( const std::string& name)
{
	std::string trans_fn= dropExtension( dropPath( _metaConfig->getTransformationFileName()));
	std::vector< std::string> incFiles;
#if _MSC_VER < 1300
	incFiles.push_back( "\""+ trans_fn + "_genops.h\"");
#else
	incFiles.push_back( "\""+ trans_fn + "_cmptempl.h\"");
#endif
	 _cg->save( name, incFiles);
}

void ConfigGenerator::saveMain( const std::string& mainFileName, const std::string& generatedRuleFileName)
{
	// print generated source
	std::ofstream source( mainFileName.c_str());
	source << timestamp( dropPath( mainFileName)) << std::endl << std::endl;
	source << "#include" << SEP << "\""<< dropPath( generatedRuleFileName) << ".h\"" << std::endl;
	source << "#include" << SEP << "\"InputFileRegistry.h\"" << std::endl;  
//	if ( false == _gzInc.empty()) {
//		source << "#include" << SEP << "\"" << _gzInc << ".h\"" << std::endl << std::endl;  
//	}
	source << "UDM_USE_DOM" << std::endl;
	source << "UDM_USE_MGA" << std::endl << std::endl;
	_main.printSource( source);
}

void ConfigGenerator::generateUCL( const std::string& currDir, GReATConfigMeta::MetaCodeGenModes mcgm)
{
	if ( !_metaConfig) {
		throw CGEx("Code generator internal error. ConfigGenerator::MetaConfig is null.");
	}
	std::string cfgPath= getPath( _cfg_fn);
	GRCWrapper().createGRC( *_metaConfig, currDir.empty() ? ( cfgPath.empty() ? getCwd() : cfgPath) : currDir, mcgm);
}

void ConfigGenerator::generateSpecCmpCode( const std::string& currDir, GReATConfigMeta::MetaCodeGenModes mcgm)
{
	if ( !_metaConfig) {
		throw CGEx("Code generator internal error. ConfigGenerator::MetaConfig is null.");
	}
	std::string cfgPath= getPath( _cfg_fn);
	GRCWrapper().createSpecCmpCode( *_metaConfig, currDir.empty() ? ( cfgPath.empty() ? getCwd() : cfgPath) : currDir, mcgm);
}

// can't use const std::string& return type, because of UDM implementation sucks.
std::string ConfigGenerator::getTransformationFileName()
{
	return _metaConfig->getTransformationFileName();
}

std::string ConfigGenerator::getOpenFileName( const std::string& fileID)
{
	return _ifrName+ ".getFileName("+ SEP+ "\""+ fileID+ "\")";
}

std::string ConfigGenerator::getCopyFileName( const std::string& fileID)
{
	return _ifrName+ ".getCopyName("+ SEP+ "\""+ fileID+ "\")";
}

std::string ConfigGenerator::getSaveFileName( const std::string& fileID, const std::string& mode)
{
	if ( "rwc" == mode)
		return getCopyFileName( fileID);
	else
		return getOpenFileName( fileID);
}

void ConfigGenerator::generateMainSkeleton( const std::vector< std::string>& fileIDs, PEComposite** tryCatchBlock)
{
	// function usage
	Function* usageFunc= new Function;
	usageFunc->_name= "usage";
	_main.add( *usageFunc);
	// function main
	Function* mainFunc= new Function;
	mainFunc->_name= "main";
	mainFunc->_returnValue= "int";
	mainFunc->_params.push_back( FuncParam_t( std::make_pair< std::string, std::string>( "int", "argc")));
	mainFunc->_params.push_back( FuncParam_t( std::make_pair< std::string, std::string>( "char*", "argv[]")));
	_main.add( *mainFunc);
	// instantiate input file registry (ifr)
	Definition* ifrDef= new Definition;
	ifrDef->_type= "InputFileRegistry";
	ifrDef->_variable= "ifr";
	mainFunc->add( *ifrDef);
	// set _ifrName here
	_ifrName= ifrDef->_variable;
	// generate usage() function
	GR::RewritingRule startRule;
	getStartRule( startRule);
	createUsageAndDefRegistration( ( std::string) startRule.name(), fileIDs, *usageFunc, *mainFunc);
	// check for '-d' default switch
	If* checkDefault= new If;
	checkDefault->appendWithAnd( "argc"+ EQEQ+ SEP+ "2");
	checkDefault->appendWithAnd("std::string( argv[ 1])"+ EQEQ+ SEP+ "\"-d\"");
	// check arguments
	If* checkArgs= new If;
	checkArgs->_expression= "argc"+ EQEQ+ SEP+ "1";
	FunctionCall* usageCall= new FunctionCall;
	usageCall->_func= usageFunc;
	checkArgs->add( *usageCall);
	checkArgs->add( *new Expression( "return -1"));
	// register arguments as input files
	Loop* argLoop= new Loop;
	argLoop->_from= "int i= 1";
	argLoop->_to= "i<argc";
	argLoop->_increment= INC+ "i";
	If* isRegOk= new If;
	isRegOk->_expression= FAL+ EQEQ+ SEP+ ifrDef->_variable+ ".registerFile( argv[ i])";
	isRegOk->add( *new Expression("throw InputFileEx( std::string( \"Invalid input file expression: \") + argv[ i])"));
	// build if 'structure'
	argLoop->add( *isRegOk);
	checkDefault->add( * new Expression("std::cout << \"Executing with default arguments.\" << std::endl"));
	checkDefault->addElse( *argLoop);
	checkArgs->addElse( *checkDefault);
	mainFunc->add( *checkArgs);
	// create try-catch block for udm exceptions.
	TryCatchBlock* udmExBlock= new TryCatchBlock;
	udmExBlock->_catchExpr= "udm_exception &e";
	udmExBlock->addCatch( *new Expression( "cout << e.what() << endl"));
	mainFunc->add( *udmExBlock);
	// main function return value
	mainFunc->add( *new Expression( "return 0"));
	*tryCatchBlock= udmExBlock;
}

void ConfigGenerator::createUsageAndDefRegistration( 
	const std::string appName, 
	const std::vector< std::string>& fileIDs, 
	PEComposite& usageComposite, 
	PEComposite& defRegComposite
	)
{
	// Get run time info, which contains necessary info to able to generate usage().
	GReATConfig::RootFolder root= GReATConfig::RootFolder::Cast( _sdnConfig->GetRootObject());
	GReATConfig::RuntimeInformation runtimeInfo= root.RuntimeInformation_child();
	GReATConfig::UsageInformation usageInfo= root.UsageInformation_child();
	if ( Udm::null == runtimeInfo)
	{
		printWarning( "No runtime information found in the configuration file. Usage function cannot be generated.");
		return;
	}
	GReATConfigRuntime runtimeConfig( runtimeInfo);
	GReATConfigUsage usageConfig( usageInfo);
	// Generate code which prints usage.
	Expression* printUsageExpr= new Expression( "std::cout << \"Usage:"+ SEP+ SEP+ appName+ ".exe");
	for ( std::vector< std::string>::const_iterator i= fileIDs.begin(); i!= fileIDs.end(); ++i)
	{
		const std::string& fileID= *i;
		std::string mode;
		_usageConfig->getMode( fileID, mode);
		// get file name belonging to 'fileID'
		std::string runtimefname= fileID+ "_INPUT_FILE.xml";
		runtimeConfig.getFileName( fileID, runtimefname);
		// copy name belonging to 'fileID'
		std::string runtimecname= fileID+ "_COPY_FILE.xml";
		runtimeConfig.getCopyName( fileID, runtimecname);
		// xsd/dtd name belonging to 'fileID'
		std::string usagexname= fileID+ "_XSD_FILE.xml";
		usageConfig.getDTDFileName( fileID, usagexname);
		// build input config expression for 'fileID'
		std::string inputConfigExpr= fileID+ EQ+ convertEscapeSequences( "\""+ runtimefname+ "\"") + 
			(mode == "rwc" ? ";"+ convertEscapeSequences( "\""+ runtimecname+ "\"")+ SEP : "") +
			(usagexname.empty()? "" : "!"+ convertEscapeSequences( "\""+ usagexname+ "\"")+ SEP);
		// add to usage expression
		printUsageExpr->_expression+= SEP+ inputConfigExpr;
		// add code to 'defRegComposite' which performs the default registration 'inputConfigExpr'.
		defRegComposite.add( *new Expression( _ifrName+ ".registerFile("+ SEP+ "\""+ inputConfigExpr+ "\")"));
	}
	// terminate usage expression and add it to usage():
	printUsageExpr->_expression+= SEP+ "or\\n\\t"+ appName+ SEP+ "-d\\n\\n\"";
	usageComposite.add( *printUsageExpr);
}

void ConfigGenerator::generateDNCode( const std::vector< std::string>& fileIDs, PEComposite& addTo)
{
	// create udm project, if needed
	bool isUdmProject= _metaConfig->isUdmProjectFile();
	if ( isUdmProject)
	{
		// datanetworks vector definition
		Definition* dnsVecDef= new Definition;
		dnsVecDef->_type= "std::vector< Udm::StaticDataNetworkSpecifier>";
		dnsVecDef->_variable= "dnsvec";
		addTo.add( *dnsVecDef);
		// set udm project info to datanetwork code printer.
		DNProjectCodePrinter::_udmProjectName= "prj";
		DNProjectCodePrinter::_dnsVecName= "dnsvec";
		CopyToMemDNCodePrinter::_udmProjectName= "prj";
		CopyToMemDNCodePrinter::_dnsVecName= "dnsvec";
	}
	// for each fileIDs...
	for( std::vector< std::string>::const_iterator i= fileIDs.begin(); i!= fileIDs.end(); ++i)
	{
		const std::string& currFileID= *i;
		std::string mode;
		_usageConfig->getMode( currFileID, mode);
		DNCodePrinter* dnCodePrinter= 0;
		CreateDNCodePrinter* codePrinterFactory= 0;
		if ( isUdmProject)
		{
			// configuration read: run in memory?
			bool runInMemory= false;
			_usageConfig->getRunInMemory( currFileID, runInMemory);
			if ( runInMemory)	// if mode= copy to memory
				codePrinterFactory= new CreateCopyToMemDnCodePrinter();
			else
				codePrinterFactory= new CreateDNProjectCodePrinter();
		}
		else
			codePrinterFactory= new CreateSmartDNCodePrinter();
		dnCodePrinter= codePrinterFactory->create( mode);
		// get neccesary information from configuration file
		std::string meta;
		_usageConfig->getMetaName( currFileID, meta);
		std::string root;
		_usageConfig->getRootClassName( currFileID, root);
		std::string dtd;
		_usageConfig->getDTDFileName( currFileID, dtd);
		// generate code...
		dnCodePrinter->generate( 
			getOpenFileName( currFileID), getSaveFileName( currFileID, mode), meta, root, "ifr.getXsdName("+SEP+ "\""+ currFileID+ "\"" + ")");
		// register the code printer with the corresponding file id.
		_dnPrinterMap[ *i]= dnCodePrinter;
	}
}

void ConfigGenerator::createOpenDNCode( PEComposite& addTo)
{
	// open code for DNs
	DNPrinterMap_t::const_iterator itDNPrinter= 0;
	for ( itDNPrinter= _dnPrinterMap.begin(); itDNPrinter!= _dnPrinterMap.end(); ++itDNPrinter)
	{
		std::string mode;
		_usageConfig->getMode( itDNPrinter->first, mode);
		addTo.add( *new CommentLine( "Open"+ SEP+ itDNPrinter->first+ SEP+ mode));
		itDNPrinter->second->getOpenCode( addTo);
	}
	if ( _metaConfig->isUdmProjectFile())
	{
		// udm project definition
		addTo.add( *new CommentLine( "Create the project"));
		Definition* udmPrjDef= new Definition;
		udmPrjDef->_type= "Udm::StaticUdmProject";
		udmPrjDef->_variable= "prj";
		udmPrjDef->addArg( "dnsvec");
		udmPrjDef->addArg( _metaConfig->getCrossLinkDiagramName()+ RES+ "diagram");
		addTo.add( *udmPrjDef);
	}
	for ( itDNPrinter= _dnPrinterMap.begin(); itDNPrinter!= _dnPrinterMap.end(); ++itDNPrinter)
	{
		itDNPrinter->second->getDnCode( addTo);
	}
}

void ConfigGenerator::createCloseDNCode( PEComposite& addTo)
{
	if ( _metaConfig->isUdmProjectFile())
	{
		// close udm project
		addTo.add( *new CommentLine( "Close the project"));
		addTo.add( *new Expression( "prj.Close()"));
	}
	for ( DNPrinterMap_t::const_iterator i= _dnPrinterMap.begin(); i!= _dnPrinterMap.end(); ++i)
	{
		std::string mode;
		_usageConfig->getMode( i->first, mode);
		addTo.add( *new CommentLine( "Close"+ SEP+ i->first+ SEP+ mode));
		i->second->getCloseCode( addTo);
	}
}

void ConfigGenerator::createPacketDefinitions( const ParamDatas_t& params, PEComposite& addTo)
{
	for( ParamDatas_t::const_iterator i= params.begin(); i!= params.end(); ++i)
	{
		Definition* paramDef= new Definition;
		paramDef->_type= PKS;
		paramDef->_variable= i->_name;
		addTo.add( *paramDef);
	}
}

void ConfigGenerator::createExternGlobalDefTemporaryRootObject( PEComposite& addTo)
{
	addTo.add( *new CommentLine("Extern global temporary root object(s)."));
	std::vector< GReATConfig::TemporaryFileUsage> tempFiles;
	_usageConfig->getTemporaryFileUsages( tempFiles);
	for( std::vector< GReATConfig::TemporaryFileUsage>::const_iterator it= tempFiles.begin(); it!= tempFiles.end(); ++it) {
		const GReATConfig::TemporaryFileUsage& currTmpFU= *it;
		std::string rootType= ( std::string)currTmpFU.MetaName() + RES+ ( std::string)currTmpFU.RootClassName();
		GlobalDefinition* rootObjectDef= new GlobalDefinition;
		rootObjectDef->_type= "extern" + SEP+ rootType;
		rootObjectDef->_variable= _backend.getTempRootDef( rootType)->_variable; //"ComesFromTranslatorCode";
		addTo.add( *rootObjectDef);
	}
}

void ConfigGenerator::createGetTemporaryRootObject( PEComposite& addTo)
{
	addTo.add( *new CommentLine("Get access to temporary root object(s)."));
	std::vector< GReATConfig::TemporaryFileUsage> tempFiles;
	_usageConfig->getTemporaryFileUsages( tempFiles);
	for( std::vector< GReATConfig::TemporaryFileUsage>::const_iterator it= tempFiles.begin(); it!= tempFiles.end(); ++it) {
		const GReATConfig::TemporaryFileUsage& currTmpFU= *it;
		const std::string& fileID= currTmpFU.MetaName(); // see createTempFileUsageInfos()! 
		const DNCodePrinter* printer= _dnPrinterMap[ fileID];
		if ( 0 == printer) {
			throw CGEx("CG internal error. Code printer not found for temporary file ID: " + fileID);
		}
		std::string dnName= printer->getDN();
		std::string rootType= ( std::string)currTmpFU.MetaName() + RES+ ( std::string)currTmpFU.RootClassName();
		Definition* rootObjectDef= new Definition;
//rootObjectDef->_type= rootType;
		rootObjectDef->_variable= _backend.getTempRootDef( rootType)->_variable; //"ComesFromTranslatorCode";
		rootObjectDef->_initValue= rootType+ RES+ "Cast("+ SEP+ lowerBegin( dnName)+ ".GetRootObject())";
		addTo.add( *rootObjectDef);
//		for (DNPrinterMap_t::const_iterator i= _dnPrinterMap.begin(); i!= _dnPrinterMap.end(); ++i) {
//			std::cout << "File ID: " << i->first << endl;
//		}
	}
}

void ConfigGenerator::createGetObjectsFromDNs( const ParamDatas_t& inputs, PEComposite& addTo)
{
	addTo.add( *new CommentLine("get objects from DNs"));
	std::set< GReATConfig::TransformationInput> transInputs;
	_usageConfig->getTransformationInputs( transInputs);
	//
	typedef std::map< std::string, std::string> RuleParamsMap_t;	// <Packets_t variable, Object variable>
	RuleParamsMap_t ruleParamsMap;
	// for each transformation input..
	PEComposite* currCode= &addTo;
	for ( std::set< GReATConfig::TransformationInput>::const_iterator itTransInputs= transInputs.begin(); itTransInputs!= transInputs.end(); ++itTransInputs)
	{
		const GReATConfig::TransformationInput& currTransInput= *itTransInputs;
		// find the corresponding dn code printer..
		std::string fileID= currTransInput.FileID();
		DNPrinterMap_t::const_iterator itFindID= _dnPrinterMap.find( fileID);
		if ( _dnPrinterMap.end() == itFindID)
			throw CGEx("FATAL: File ID: "+ fileID+ "not exist in the file usage block of configuration file.");
		// file id exists here. 
		// generate root object definition:
		std::string meta;
		_usageConfig->getMetaName( fileID, meta);
		std::string root;
		_usageConfig->getRootClassName( fileID, root);
		std::string dnName= itFindID->second->getDN();
		Definition* rootObjectDef= new Definition;
		rootObjectDef->_type= meta+ RES+ root;
		rootObjectDef->_variable= PE::createUniqueName( "root"+ meta);
		rootObjectDef->_initValue= meta+ RES+ root+ RES+ "Cast("+ SEP+ lowerBegin( dnName)+ ".GetRootObject())";
		currCode->add( *rootObjectDef);
		// traverse the transformation input object path
		std::vector< std::string> objectPath= currTransInput.ObjectPath();
		if ( objectPath.empty())
			throw CGEx("FATAL: Empty object path found in transformation input of configuration file.");
		// root object in object path:
		const string& objectPathElem = *(objectPath.begin());
		std::string objectType;
		std::vector< std::pair< std::string, std::string> > objectAttrs;
		_usageConfig->parseObjectPathElem( objectPathElem, objectType, objectAttrs);
		if ( objectType != root) // type mismatch..
			throw CGEx("FATAL: Root object type mismatch error. Root in usage info: "+ objectType+ " Root in meta info: "+ root);
		// remove root from opject path
		objectPath.erase( objectPath.begin());
		// bind objects to start rule input arguments
		std::string inputPackets;
		for( ParamDatas_t::const_iterator itInputs= inputs.begin(); itInputs!= inputs.end(); ++itInputs)
		{
			if ( (std::string) itInputs->_param.name() == (std::string) currTransInput.PortName())
			{
				inputPackets= itInputs->_name;	// found
				break;
			}
		}
		// 'object selection based on attribute matching' & 'children traversal' code generation:
		If* checkAttr= NULL;
		std::string varName= rootObjectDef->_variable;
		if ( generateCheckRootAttributes( rootObjectDef->_variable, objectAttrs, *currCode, &checkAttr))
		{
			// there were root attributes specified in object path
			checkAttr->add( *new Expression("std::cout << \"Root not selected, empty transformation!\""));
			currCode= &checkAttr->getElseBranch();
			varName= rootObjectDef->_variable;
//			if ( objectPath.empty()) // only root object in object path
//				checkAttr->addElse( *new Expression( lowerBegin(inputPackets)+ ".push_back("+ SEP+ lowerBegin( rootObjectDef->_variable)+ ")"));
//			else // root object + child(ren) in object path
//				currCode= generateObjectTraversal( rootObjectDef->_variable, meta, objectPath, &checkAttr->getElseBranch());
		}
		if ( !objectPath.empty()) // root object + child(ren) in object path
			currCode= generateObjectTraversal( rootObjectDef->_variable, meta, objectPath, currCode, varName);
		ruleParamsMap[ lowerBegin(inputPackets)]= varName;
	}
	currCode->add( *new CommentLine("Add objects to input packets."));
	for( RuleParamsMap_t::const_iterator itParam= ruleParamsMap.begin(); itParam != ruleParamsMap.end(); ++itParam) {
		currCode->add( *new Expression( itParam->first+ ".push_back("+ SEP+ itParam->second+ ")"));
	}
}

// this function is called recursively.
PEComposite* ConfigGenerator::generateObjectTraversal( 
	const std::string& parent,
	const std::string& meta,
	std::vector< std::string>& objectPath, 
	PEComposite* addTo,
	std::string& variableName
	)
{
		if ( objectPath.empty())
			return addTo;	// do nothing
		// generate code for 'get children'
		Definition* childrenDef= new Definition;
		string objectPathElem = *(objectPath.begin());
		std::string objectType;
		std::vector< std::pair< std::string, std::string> > objectArgs;
		_usageConfig->parseObjectPathElem( objectPathElem, objectType, objectArgs);
		childrenDef->_type= "std::set<"+ SEP+ meta+ RES+ objectType+ ">";
		childrenDef->_variable= PE::createUniqueName( lowerBegin( objectType+ "s"));
		childrenDef->_initValue= parent+ "."+ objectType+ "_kind_children()";
		addTo->add( *childrenDef);
		objectPath.erase( objectPath.begin());
		// generate code which traverse children:
		// 1, if there are attributes specified in object path -OR-
		// 2, objectPath is not empty (traverse the children of current objects)
//		if ( !objectArgs.empty() || !objectPath.empty())
//		{
			// generate code which traverse children
			Loop* traverseChildrenLoop= new Loop;
			traverseChildrenLoop->createIteratorLoop( IT+childrenDef->_variable, childrenDef->_variable, childrenDef->_type);
			addTo->add( *traverseChildrenLoop);
			addTo= traverseChildrenLoop;
			Definition* currChildDef= new Definition;
			currChildDef->_type= CST+ SEP+ meta+ RES+ objectType+ REF;
			currChildDef->_variable= PE::createUniqueName( CURR+ objectType);
			currChildDef->_initValue= PTR+ traverseChildrenLoop->_loopVariable;
			addTo->add( *currChildDef);
			variableName= currChildDef->_variable;
			if ( !objectArgs.empty())
			{
				// attributes for current object... 
				generateCheckChildrenAttributes( currChildDef->_variable, objectArgs, *addTo);
//				if ( objectPath.empty())
//					addTo->add( *new Expression( lowerBegin( inputPackets)+ ".push_back("+ SEP+ currChildDef->_variable+ ")"));
			}
			if ( !objectPath.empty())
			{
				// still objects in object path -> recursion
				return generateObjectTraversal( currChildDef->_variable, meta, objectPath, addTo, variableName);
			}
//		}
//		else
//		{
//			// generate code which inserts objects into packets
//			addTo->add( *new Expression( lowerBegin( inputPackets)+ ".insert("+ SEP+ lowerBegin( inputPackets)+ ".end(),"+ SEP+
//				childrenDef->_variable+ ".begin(),"+ SEP+ childrenDef->_variable+ ".end())" ));
//		}
		return addTo;
}

void ConfigGenerator::createExecuteTransformation( const Class& ruleClass, const ParamDatas_t& ruleInputs, const ParamDatas_t& ruleOutputs, PEComposite& addTo)
{
	addTo.add( *new CommentLine( "transformation"));
	FuncOperator* funcOp= new FuncOperator;
	funcOp->_class= &ruleClass;
	ParamDatas_t::const_iterator i;
	for( i= ruleInputs.begin(); i!= ruleInputs.end(); ++i)
		funcOp->_params.push_back( i->_name);
	for( i= ruleOutputs.begin(); i!= ruleOutputs.end(); ++i)
		funcOp->_params.push_back( i->_name);
	addTo.add( *funcOp);
}

void ConfigGenerator::getStartRule( GR::RewritingRule& startRule)
{
	GR::RootFolder grRootFolder;
	_metaConfig->getRootFolder( grRootFolder, getPath( _cfg_fn));
	_usageConfig->getStartRule( grRootFolder, startRule);
}

int ConfigGenerator::getNoOfArgs( const std::vector< std::string>& fileIDs)
{
	int argc= 0;
	for( std::vector< std::string>::const_iterator i= fileIDs.begin(); i!= fileIDs.end(); ++i)
	{
		std::string mode;
		if ( false== _usageConfig->getMode( *i, mode))
			throw CGEx("FATAL: Invalid file id: "+ *i);
		if ( mode == "rwc")
			argc+= 2;	// read-write copy has two file names. (open & saveas)
		else
			++argc;
	}
	return argc;
}

bool ConfigGenerator::generateCheckRootAttributes( 
	const std::string& objectName, 
	const std::vector< std::pair< std::string, std::string> >& objectAttrs,
	PEComposite& addTo,
	If** checkAttr
)
{
	return 0 != ( *checkAttr = generateCheckAttributes( objectName, objectAttrs, addTo));
}

bool ConfigGenerator::generateCheckChildrenAttributes( 
	const std::string& objectName, 
	const std::vector< std::pair< std::string, std::string> >& objectAttrs,
	PEComposite& addTo
)
{
	If* checkAttr= 0;
	if ( 0!= ( checkAttr= generateCheckAttributes( objectName, objectAttrs, addTo)))
		checkAttr->add( *new Expression("continue"));
	return 0 != checkAttr;
}

If* ConfigGenerator::generateCheckAttributes( 
	const std::string& objectName, 
	const std::vector< std::pair< std::string, std::string> >& objectAttrs,
	PEComposite& addTo
	)
{
	if ( objectAttrs.empty())
		return 0;
	If* checkAttr= new If;
	for( std::vector< std::pair< std::string, std::string> >::const_iterator i= objectAttrs.begin(); i!= objectAttrs.end(); ++i)
	{
		const std::string& attrName= i->first;
		const std::string& attrValue= i->second;
		std::string::size_type findPoint;
		if ( !attrValue.empty() && ( attrValue[ 0]== '\"') && ( attrValue[ attrValue.size()-1]== '\"'))
		{
			// string
			Definition* stringDef= new Definition;
			stringDef->_type= "std::string";
			stringDef->_variable= PE::createUniqueName( objectName+ "_"+ attrName);
			addTo.add( *stringDef);
			addTo.add( *new Expression( objectName+ ".GetStrValue("+ SEP+ "\""+ attrName+ "\","+ SEP+ stringDef->_variable+ ")"));
			checkAttr->appendWithOr( stringDef->_variable+ NEQ+ SEP+ attrValue);
		}
		else if ( std::string::npos != ( findPoint= attrValue.find( '.')))
		{
			// real
			for ( std::string::const_iterator itValue= attrValue.begin(); itValue!= attrValue.end(); ++itValue)
			{
				if ( (false== isdigit( *itValue)) && (*itValue!= '.'))
				{
					printError( "Invalid real attribute value in object path: "+ attrName+ EQ+ SEP+ attrValue+ SEP+ "???");
					break;
				}
			}
			Definition* realDef= new Definition;
			realDef->_type= "double";
			realDef->_variable= PE::createUniqueName( objectName+ "_"+ attrName);
			addTo.add( *realDef);
			addTo.add( *new Expression( objectName+ ".GetRealValue("+ SEP+ "\""+ attrName+ "\","+ SEP+ realDef->_variable+ ")"));
			checkAttr->appendWithOr( realDef->_variable+ NEQ+ SEP+ attrValue);
		}
		else 
		{
			// integer
			for ( std::string::const_iterator itValue= attrValue.begin(); itValue!= attrValue.end(); ++itValue)
			{
				if ( false== isdigit( *itValue))
				{
					printError( "Invalid integer attribute value in object path: "+ attrName+ EQ+ SEP+ attrValue+ SEP+ "???");
					break;
				}
			}
			Definition* intDef= new Definition;
			intDef->_type= "int";
			intDef->_variable= PE::createUniqueName( objectName+ "_"+ attrName);
			addTo.add( *intDef);
			addTo.add( *new Expression( objectName+ ".GetIntValue("+ SEP+ "\""+ attrName+ "\","+ SEP+ intDef->_variable+ ")"));
			checkAttr->appendWithOr( intDef->_variable+ NEQ+ SEP+ attrValue);
		}

	}
	addTo.add( *checkAttr);
	return checkAttr;
}

class TempFileUsageConverter
{
public:
	TempFileUsageConverter( GReATConfigUsage* usageConfig) : _usageConfig( usageConfig) {}
	void operator()( const GReATConfig::TemporaryFileUsage& tfu) {
		_usageConfig->addFileUsage( tfu, tfu.MetaName(), false, "w");
	}
private:
	GReATConfigUsage* _usageConfig;
};

void ConfigGenerator::createTempFileUsageInfos()
{
	std::vector< GReATConfig::TemporaryFileUsage> tempFiles;
	_usageConfig->getTemporaryFileUsages( tempFiles);
	std::for_each( tempFiles.begin(), tempFiles.end(), TempFileUsageConverter( _usageConfig));
}
