#if !defined(GRDBackEnd_h)
#define GRDBackEnd_h

#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>

#include <map>
#include "IGRDBackEnd.h"

namespace GRDI
{
typedef boost::mutex::scoped_lock Lock;
typedef std::map<LocationID, IFunctionPtr> LocationIDIFunctionPtrMap;

class GRDBackEnd : public IGRDBackEnd,
                   public DebugInterface
{
  EventHook*          m_eventHook;

  EventSet            m_eventSet;

  bool                m_suspended;

  boost::condition    m_resume;
  boost::mutex        m_mutexCtrl;
  boost::mutex        m_mutexSuspend;
  boost::mutex        m_mutexWatch;

  IFunctionPtr        m_startFunc;
  IFunctionPtrSet     m_rootFuncs;
  LocationIDVec       m_rootLocs;
  FunctionDescrMap    m_functions;
  LocIDLocMap         m_id2loc;
  LocLocIDMap         m_loc2id;
  LocationIDIFunctionPtrMap m_refMap;

  LocationIDSet       m_breakpoints;

  Context             m_context;

  LocationID          m_locID;

  ContextID           m_popEventCtxID;

  bool                m_tempBreakpoint;
  bool                m_debugInfo;

  std::string         m_programFileName;
  std::string         m_workingDirectory;
  std::string         m_exceptionDescr;
  int                 m_popCtxNum;

  SynchType           m_synchStatus;

  protected:

    void resetEventSet();
    void resetFlags();
    void resetContext();
    void addNewContext(const FunctionCallDescr& funcCallDescr, const LocationID& currLocationID);
    void refreshCurrContext(const FunctionCallDescr& funcCallDescr, const LocationID& currLocationID);
    ContextDescr&  popContext();

    LocationID addLocation(const LocationID& parLocID, const Location& loc);

    void orderCalls();
    void setLocations(const LocationID& parLocID, const IFunctionPtr& func);

    void orderFunctionCalls(FunctionDescr& funcDescr);

    LocationID findBeginLocation(const IFunctionPtr& func);
    LocationID findEndLocation(const IFunctionPtr& func);
    LocationID findCallLocation(const IFunctionPtr& func);
    LocationID findLocation(const FunctionID& id, const LocationType& type); 
    void checkBreakpoints(const LocationID& id);

  public:

    GRDBackEnd();
      
    virtual ~GRDBackEnd() {}

    
    // =========== IGRDBackEnd
    void stepIn(const FunctionCallDescr& funcValDescr);
    void step(const FunctionCallDescr& funcValDescr);
    void stepOut(const FunctionCallDescr& funcValDescr);

    DebugInterface* getDebugInterface();

    void setStartFunction(const IFunctionPtr&);
    void addRootFunction(const IFunctionPtr&);
    void registerFunctionCall(const FunctionDescr& funcDescr);
    void clearSource();

    bool synchronize();
    SynchType  getSynchStatus();
    bool  debugInfo() {return m_debugInfo;}
    const std::string& getProgramFileName();
    const std::string& getWD();
    void saved();
    void loaded();
    void closed();
    void stopped();
    void exception(const std::string& str);
    void up();
    void down();
    void start();
    void finish();
    void reportProgress(double prg);


  public:

    // =========== DebugInterface
    Error getStatus(Status& status);
    Error getContextStacktSize(int& count);
    Error getCurrentContextID(ContextID& context);
    Error getContextDescr(const ContextID& contextID, ContextDescr& contextDescr);
    Error getCallerContext(const  ContextID& called,
								                  ContextID& caller);

   	Error getCurrentLocation( const  ContextID& context,
 									                   LocationID& func);

   	Error getLastException( std::string& str);


    Error notifyContextPop(const ContextID& context); 

    Error  suspend();
    Error  resume();
    Error  execute();
    Error  run();
    Error  stop();
    Error  pause();
    Error  quit();
    Error  save();
    Error  load(const std::string&);
    Error  close();

    Error setBreakpoint(const LocationID& loc);
    Error clearBreakpoint(const LocationID& loc);
    Error clearAllBreakpoints();
    Error getBreakPoints(LocationIDSet& breakpoints);

    
    Error getRootLocations(LocationIDVec& rootLocs);

    Error getFunctionDescr(const IFunctionPtr& func,
                                 FunctionDescr& descr);

    Error getLocationFromID(const LocationID& locationID,
                                      Location& location);


    Error setEventHook(EventHook* hook);


};
};//namespace GRDI
#endif